/* diskdmp -- Dump JV3 dsk files to real TRS-80 Model 3/4 disks */
/* 1998 A.R.Duell and placed under the GPL */

#include<stdio.h>
#include<string.h>
#include<unistd.h>
#include<fcntl.h>
#include<linux/fd.h>
#include<linux/fdreg.h>
#include<sys/ioctl.h>

#define HDR_COUNT 2901 
#define BYTE_COUNT 256

/* bit allocations for the flags word in a header */
#define DENSITY 0x80
#define LOCKED 0x40
#define DELDAM 0x20
#define SIDE 0x10
#define ERROR 0x08
#define SHORT 0x04
#define SIZE 0x03

/* Floppy controller data rate selection codes */
#define RATE250 2
#define RATE300 1

/* 2 useful FDC commands that are not defined in fdreg.h */
#define FD_DD_WRITE 0x45
#define FD_DD_DEL_WRITE 0x49

typedef struct {int track; /* track number */
                int sector; /* sector number */
                int flags; /* sector type flags */
               } hdr_t;

hdr_t headers[HDR_COUNT]; /* array of sector headers from the .dsk file */

int read_headers(FILE *dsk_file)
  {
    int min_sec=254; /* lowest sector on a track */
    int max_sec=0; /* highest sector on a track */
    int min_trk=254; /* lowest track number */
    int max_trk=0; /* highest track number */
    int dbl_side=0; /* set to 1 if this disk is doublesided */
    int dir_trk; /* highest number track written with deleted DAMs */
    int last_sector=0; /* higest sector number with valid data */
    int hdr; /* header index */
    int sngl_den=0; /* set if single density sector found */
    int illegal=0; /* set if illegal flags found */
    int wrprot; /* write protect flag from the dsk file */

    for(hdr=0;hdr<HDR_COUNT;hdr++)
      {
        /* read next header */
        headers[hdr].track=getc(dsk_file);
        headers[hdr].sector=getc(dsk_file);
        headers[hdr].flags=getc(dsk_file);
        if ((headers[hdr].track==0xFF) && (headers[hdr].sector==0xFF))
          {
            /* skip if this is an unused header */
            continue;
          }
        min_trk=(headers[hdr].track<min_trk)?headers[hdr].track:min_trk;
        max_trk=(headers[hdr].track>max_trk)?headers[hdr].track:max_trk;
        min_sec=(headers[hdr].sector<min_sec)?headers[hdr].sector:min_sec;
        max_sec=(headers[hdr].sector>max_sec)?headers[hdr].sector:max_sec;
        dbl_side=(headers[hdr].flags&SIDE)?1:dbl_side;
        dir_trk=(headers[hdr].flags&DELDAM)?headers[hdr].track:dir_trk;
        sngl_den=(headers[hdr].flags&DENSITY)?sngl_den:1;
        illegal=(headers[hdr].flags&(LOCKED|ERROR|SHORT|SIZE))?1:illegal;
      }
    for(hdr=HDR_COUNT-1;hdr>=0;hdr--)
      {
        if ((headers[hdr].track!=0xFF) || (headers[hdr].sector!=0xFF))
          {
            last_sector=hdr;
            break;
           }
       }
    wrprot=getc(dsk_file);
    printf("lowest track = %d, highest track = %d\n",min_trk,max_trk);
    printf("lowest sector = %d, highest sector =%d\n",min_sec,max_sec);
    printf("directory found on track %d\n",dir_trk);
    printf(dbl_side ? "double sided\n" : "single sided\n");
    printf(wrprot ? "" : "write protected\n");
    if (sngl_den)
      {
        printf("This image contains single density sectors\n");
        exit(1);
      }
    if (illegal)
      {
        printf("This image contains illegal flag settings\n");
        exit(1);
      }
    return(last_sector+1);
  }

int data_rate(int device)
  {
    struct floppy_drive_params params;
    int rate;

    ioctl(device,FDGETDRVPRM,&params);
    rate=(params.rps==6)?RATE300:RATE250;
    printf("rps=%d,rate=%d\n",params.rps,rate);
    return (rate);
  } 
  
void read_block(FILE *dsk_file,unsigned char *data)
  {
     int i;
     for(i=0;i<BYTE_COUNT;i++)
       {
         data[i]=getc(dsk_file);
       }
   }

void recalibrate(int device, int rate)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;
    cmd.length=0;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_RECALIBRATE;
    cmd.cmd[1]=0;
    cmd.cmd_count=2;
    if(ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on recalibrate\n");
         exit(1);
      }
  }

void seek(int device, int rate, int track, int dblstep)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;
    cmd.length=0;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_SEEK;
    cmd.cmd[1]=0;
    cmd.cmd[2]=track*(1<<dblstep);
    cmd.cmd_count=3;
    if(ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on seek to track %d\n",track);
         exit(1);
      }
  }

void write_sec(int device,int rate,int track,int sector,int deldam,int side, 
               unsigned char *data)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=data;
    cmd.length=BYTE_COUNT;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR | FD_RAW_WRITE;
    cmd.cmd[0]=deldam?FD_DD_DEL_WRITE:FD_DD_WRITE;
    cmd.cmd[1]=side?4:0;
    cmd.cmd[2]=track;
    cmd.cmd[3]=side?1:0;
    cmd.cmd[4]=sector;
    cmd.cmd[5]=1; /* 256 byte MFM sectors */
    cmd.cmd[6]=19; /* last sector + a bit */
    cmd.cmd[7]=14; /* gap length */
    cmd.cmd[8]=0xFF; /* 256 bytes, shouldn't be needed */
    cmd.cmd_count=9;
    if (ioctl(device,FDRAWCMD,&cmd)<0)
      {
        fprintf(stderr,"Error writing sector %d, track %d\n",sector,track);
        exit(1);
      }
   if (cmd.reply[1]&4) 
     {
       fprintf(stderr,"Sector %d, Track %d not found\n",sector, track);
     }
   if (cmd.reply[1]&1)
     {
       fprintf(stderr,"Missing AM, sector %d, track %d\n",sector,track);
     }
  }

void write_disk(int device, FILE *dsk_file, int rate, int dblstep, int blocks)
  {
    unsigned char data[BYTE_COUNT];
    int this_block;
    int this_track=-1;

    recalibrate(device,rate);
    for(this_block=0;this_block<blocks;this_block++)
      {
        read_block(dsk_file,data);
        if ((headers[this_block].track==0xFF) 
            && (headers[this_block].sector==0xFF))
          {
            /* skip if this is an unused header */
            continue;
          }
        if(headers[this_block].track!=this_track)
          {
            this_track=headers[this_block].track;
            seek(device,rate,this_track,dblstep);
          } 
        write_sec(device,rate,headers[this_block].track,
              headers[this_block].sector,
              headers[this_block].flags&DELDAM,
              headers[this_block].flags&SIDE,data);
      }
  }
void usage(void)
  {
    printf("usage : diskdmp (-d) (-w device) dskfile\n\n");
    printf(" -d - double step floppy drive\n");
    printf(" -w device - write dsk file to real floppy in device \n");
    printf("             (otherwise just print track and sector usage)\n\n");
    printf("This program is released under the GPL v2.0. See 'copying' for\n");
    printf("further details\n");
    exit(1);
  }

int main(int argc,char **argv)
  {
    int option; /* Current option character */
    int writing=0; /* dsk file is to be written to real disk */
    int dblstep=0; /* double step the floppy drive */
    int block_count; /* last block used + 1*/
    int rate; /* data rate code */
    char device_name[128];
    FILE *dsk_file; 
    int device;

    optind = 1;
    while ((option=getopt(argc,argv,"dw:?"))!=-1)
      {
        switch(option)
          {
            case 'd' : dblstep=1;
                       break;
            case 'w' : writing=1;
                       strcpy(device_name,optarg);
                       break;
            case '?' : usage();
          }
      }
    if (optind != argc-1) {usage();} /* Error if incorrect number of files */
    if (!(dsk_file=fopen(argv[optind],"r")))
      { 
        printf("Error opening dsk file\n");
        exit(1);
      }
    block_count=read_headers(dsk_file);  
    printf("Block count = %d\n",block_count);
    if (writing)
      {
        if ((device=open(device_name,O_RDWR,0))==-1)
          {
            printf("Error opening device %s\n",device_name);
            exit(1);
          }
        rate=data_rate(device);
        write_disk(device,dsk_file,rate,dblstep,block_count);
        close(device);
      }
    fclose(dsk_file);
    exit(0);
  }
